(************************************************************************
  Author:	Robert Rossmair
		Wagelwitz, Rodaer Str. 6
		04688 Mutzschen
		Germany

	e-mail:	Robert.Rossmair@t-online.de

  Module:	NumEdit

  Version:	03-OCT-1997

  Description:	Number editors

      Classes:	TIntEdit	(integer)
		TFloatEdit	(floating point)

  Copyright (C) 1997  Robert Rossmair

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program (file COPYING); if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************************************************************)
unit NumEdit;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Forms,
  Controls, StdCtrls, ExtCtrls, Consts, Buttons;

const
  FixPointUnit = 1 shl 16;
  MaxFixed = High(Longint) / FixPointUnit;
  MinFixed = Low(Smallint);
type
  TIntEdit = class(TCustomEdit)
  private
    { Private declarations }
    FMin: Longint;
    FMax: Longint;
    FValidateOnExit: Boolean;
    procedure DoExit; override;
    procedure Error;
    procedure SetMin(NewMin: Longint);
    procedure SetMax(NewMax: Longint);
    function GetValue: Longint;
    procedure SetValue(NewValue: Longint);
  protected
    procedure Change; override;
    procedure KeyPress(var Key: Char); override;
    property MaxLength default 11;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Validate;
    property Text;
  published
    { Published declarations }
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
{ When the Control resides in a modal dialog,
  validation should happen only when quitting
  the dialog by "OK"; no validation should
  be performed if the dialog is canceled.
  Validation can be switched off by setting
  ValidateOnExit to False. }
    property ValidateOnExit: Boolean read FValidateOnExit write FValidateOnExit default True;
    property Min: LongInt read FMin write SetMin default $80000000;
    property Max: LongInt read FMax write SetMax default MaxLongint;
    property Value: Longint read GetValue write SetValue default 0;
  end;

  TDigits = 0..18;
  TFloatEdit = class(TCustomEdit)
  private
    { Private declarations }
    FScrollBar: TScrollBar;
    FMin: Extended;
    FMax: Extended;
    FPrecision: TDigits;
    FDigits: TDigits;
    FFormat: TFloatFormat;
    FScrollBarChanging:	Boolean;
    FChangingByScrollBar: Boolean;
    FValidateOnExit: Boolean;
    procedure DoExit; override;
    procedure Error;
    procedure SetMin(NewMin: Extended);
    procedure SetMax(NewMax: Extended);
    function GetFixPoint: TFixed;
    function GetRelativeValue: Extended;
    function GetValue: Extended;
    procedure SetPrecision(APrecision: TDigits);
    procedure SetDigits(AValue: TDigits);
    procedure SetFormat(AFormat: TFloatFormat);
    procedure SetFixPoint(NewValue: TFixed);
    procedure SetValue(NewValue: Extended);
    procedure SetScrollBar(AScrollBar: TScrollBar);
  protected
    procedure Change; override;
    procedure KeyPress(var Key: Char); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Validate;
    property FixPoint: TFixed read GetFixPoint write SetFixPoint;
    property Text;
  published
    { Published declarations }
    procedure ScrollBarChange(Sender: TObject);
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
{ When the Control resides in a modal dialog,
  validation should happen only when quitting
  the dialog by "OK"; no validation should
  be performed if the dialog is canceled.
  Validation can be switched off by setting
  ValidateOnExit to False. }
    property ValidateOnExit: Boolean read FValidateOnExit write FValidateOnExit default True;
    property Precision: TDigits read FPrecision write SetPrecision default 18;
    property Digits: TDigits read FDigits write SetDigits default 5;
    property Format: TFloatFormat read FFormat write SetFormat default ffFixed;
    property Min: Extended read FMin write SetMin;
    property Max: Extended read FMax write SetMax;
    property Value: Extended read GetValue write SetValue;
    property ScrollBar: TScrollBar read FScrollBar write SetScrollBar;
  end;

procedure Register;

implementation

resourcestring
  SFloatOutOfRange = 'Value must be between %g and %g';

constructor TIntEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  MaxLength := 11;
  FMin := $80000000;
  FMax := MaxLongint;
  Text := '0';
  FValidateOnExit := True;
end;

procedure TIntEdit.Change;
var
  X: Integer;
begin
  try
    X := GetValue;
    if (X >= FMin) and (X <= FMax) then inherited;
  except
    on EConvertError do;
  end;
end;

procedure TIntEdit.DoExit;
begin
  if FValidateOnExit and not (csDesigning in ComponentState)
    then Validate;
  inherited DoExit;
end;

procedure TIntEdit.Error;
var
  StackTop: record end;
  Stack: record
    BPorEBP: Integer;
    ReturnAddress: Pointer;
  end absolute StackTop;
begin
  raise ERangeError.CreateFmt(SOutOfRange, [FMin, Fmax]) at Stack.ReturnAddress;
end;

procedure TIntEdit.Validate;
var
  X: Longint;
begin
  try
    X := StrToInt(Text);
    if (X < FMin) or (X > FMax) then Error;
  except
    (Owner as TForm).ActiveControl := Self;
    raise
  end;
end;

procedure TIntEdit.KeyPress(var Key: Char);
begin
  case Key of
    #0..#32,
    '-', '0'..'9': exit;
  end;
  Key := #0;
end;

procedure TIntEdit.SetMin(NewMin: Longint);
begin
  FMin := NewMin;
  if FMax < NewMin then FMax := NewMin;
  if GetValue < NewMin then SetValue(NewMin);
end;

procedure TIntEdit.SetMax(NewMax: Longint);
begin
  FMax := NewMax;
  if FMin > NewMax then FMin := NewMax;
  if GetValue > NewMax then SetValue(NewMax);
end;

function TIntEdit.GetValue: Longint;
begin
  Result := StrToInt(Text);
end;

procedure TIntEdit.SetValue(NewValue: Longint);
begin
  Text := IntToStr(NewValue);
end;

constructor TFloatEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPrecision := 18;
  FDigits := 5;
  FFormat := ffFixed;
  Text := '0';
  FValidateOnExit := True;
end;

procedure TFloatEdit.DoExit;
begin
  if FValidateOnExit and not (csDesigning in ComponentState)
    then Validate;
  inherited DoExit;
end;

procedure TFloatEdit.Error;
var
  StackTop: record end;
  Stack: record
    BPorEBP: Integer;
    ReturnAddress: Pointer;
  end absolute StackTop;
begin
  raise ERangeError.CreateFmt(SFloatOutOfRange, [FMin, Fmax]) at Stack.ReturnAddress;
end;

procedure TFloatEdit.Validate;
var
  X: Extended;
begin
  try
    X := StrToFloat(Text);
    if (X < FMin) or (X > FMax) then Error;
  except
    (Owner as TForm).ActiveControl := Self;
    raise
  end;
end;

procedure TFloatEdit.KeyPress(var Key: Char);
begin
  case Key of
    #0..#32,
    '-', '+', '0'..'9', 'E', 'e': exit;
  else
    if Key = DecimalSeparator then exit;
  end;
  Key := #0;
end;

function TFloatEdit.GetFixPoint: TFixed;
begin
  Longint(Result) := Round(FixPointUnit * StrToFloat(Text));
end;

procedure TFloatEdit.SetFixPoint(NewValue: TFixed);
begin
  Text := FloatToStrF(Longint(NewValue)/FixPointUnit, ffFixed, FPrecision, FDigits);
  if FMax > MaxFixed then FMax := MaxFixed;
  if FMin < MinFixed then FMin := MinFixed;
end;

procedure TFloatEdit.SetPrecision(APrecision: TDigits);
begin
  if APrecision <> FPrecision then
  begin
    FPrecision := APrecision;
    SetValue(Value);
  end;
end;

procedure TFloatEdit.SetDigits(AValue: TDigits);
begin
  if AValue <> FDigits then
  begin
    FDigits := AValue;
    SetValue(Value);
  end;
end;

procedure TFloatEdit.SetFormat(AFormat: TFloatFormat);
begin
  if AFormat <> FFormat then
  begin
    FFormat := AFormat;
    SetValue(Value);
  end;
end;

procedure TFloatEdit.SetMin(NewMin: Extended);
begin
  FMin := NewMin;
  if not (csReadingState in ControlState) then
  begin
    if FMax < NewMin then FMax := NewMin;
    if GetValue < NewMin then SetValue(NewMin);
  end;
end;

procedure TFloatEdit.SetMax(NewMax: Extended);
begin
  FMax := NewMax;
  if not (csReadingState in ControlState) then
  begin
    if FMin > NewMax then FMin := NewMax;
    if GetValue > NewMax then SetValue(NewMax);
  end;
end;

function TFloatEdit.GetValue: Extended;
begin
  Result := StrToFloat(Text);
end;

procedure TFloatEdit.SetValue(NewValue: Extended);
begin
  Text := FloatToStrF(NewValue, ffFixed, FPrecision, FDigits);
end;

function TFloatEdit.GetRelativeValue: Extended;
var
  X: extended;
begin
  Result := 0;
  if Max > Min then
  try
    X := GetValue;
    if X < FMin then exit;
    if X >= FMax
      then Result := 1.0
      else Result := (X-FMin)/(FMax-FMin);
  except
    on EConvertError do;
  end;
end;

procedure TFloatEdit.SetScrollBar(AScrollBar: TScrollBar);
begin
  if AScrollBar <> nil then
  begin
    AScrollBar.SetParams(Round(High(Smallint) * GetRelativeValue), 0, High(Smallint));
    AScrollBar.LargeChange := 2048;
    AScrollBar.OnChange := ScrollBarChange;
  end;
  FScrollBar := AScrollBar;
end;

procedure TFloatEdit.ScrollBarChange(Sender: TObject);
begin
  if not FScrollBarChanging then
  if Sender = FScrollBar then
  begin
    FChangingByScrollBar := True;
    try
      if FScrollBar.Max <> FScrollBar.Min then
	Value := Min + FScrollBar.Position * (Max-Min)/(FScrollBar.Max-FScrollBar.Min);
    finally
      FChangingByScrollBar := False;
    end;
  end;
end;

procedure TFloatEdit.Change;
var
  X: Extended;
begin
  try
    X := GetValue;
    if (X < FMin) or (X > FMax) then Exit;
    if not FChangingByScrollBar then
    if FScrollBar <> nil then
    begin
      FScrollBarChanging := True;
      try
	FScrollBar.Position := FScrollBar.Min + Round((FScrollBar.Max-FScrollBar.Min) * GetRelativeValue);
      finally
	FScrollBarChanging := False;
      end;
    end;
    inherited Change;
  except
    on EConvertError do;
  end;
end;

const
{$IFDEF Bilch}
  PalettePage = 'Homebrew';
{$ELSE}
  PalettePage = '3rdParty';
{$ENDIF}
procedure Register;
begin
  RegisterComponents(PalettePage, [
	TIntEdit,
	TFloatEdit]);
end;

end.
